from typing import Optional

from .Data import Data
from .Base import Base

from .RC import PortObjPos


class NodePort(Base):
    """Base class for inputs and outputs of nodes"""

    def __init__(self, node, io_pos, type_, label_str):
        Base.__init__(self)

        self.node = node
        self.io_pos = io_pos
        self.type_ = type_
        self.label_str = label_str

    def data(self) -> dict:
        d = super().data()
        d.update({
            'type': self.type_,
            'label': self.label_str,
        })
        return d


class NodeInput(NodePort):

    def __init__(self, node, type_, label_str='', add_data=None):
        super().__init__(node, PortObjPos.INPUT, type_, label_str)

        # data can be used to store additional data for enhanced data input ports
        self.add_data = add_data

    def data(self) -> dict:
        d = super().data()
        return d


class NodeOutput(NodePort):
    def __init__(self, node, type_, label_str=''):
        super().__init__(node, PortObjPos.OUTPUT, type_, label_str)

        self.val: Optional[Data] = None

    # def data(self) -> dict:
    #     data = super().data()
    #
    #     data['val'] = self.val if self.val is None else self.val.get_data()
    #
    #     return data



"""
This module defines features and behavior that applies to most internal
components. It defines the Base (parent) class.
"""


def complete_data(data: dict) -> dict:
    """
    Default implementation for supplementing data with additional (e.g. frontend)
    information. For example, a frontend can store additional information like
    current position, color of a node, etc. by replacing this function by an
    according custom handler.
    """
    return data


class Event:
    def __init__(self, *args):
        self.args = args
        self._slots = []

    def connect(self, callback):
        """
        Registers a callback function. The callback must accept compatible arguments.
        """
        self._slots.append(callback)

    def disconnect(self, callback):
        """
        De-registers a callback function. The function must have been added previously.
        """
        self._slots.remove(callback)

    def emit(self, *args):
        """
        Emits an event by calling all registered callback functions in the order they
        were registered, with parameters given by *args.
        """

        # I am assuming that the for-each loop keeps the overhead small in case
        # there are no slots registered, but one might want to profile that.

        for cb in self._slots:
            cb(*args)

class IDCtr:
    """
    A simple ascending integer ID counter.
    Guarantees uniqueness only during lifetime.
    This approach is preferred over UUIDs because UUIDs require system support
    which might not be available everywhere.
    """

    def __init__(self):
        self.ctr = -1

    def count(self):
        """increases the counter and returns the new count. first time is 0"""
        self.ctr += 1
        return self.ctr

    def set_count(self, cnt):
        if cnt < self.ctr:
            raise Exception("Decreasing ID counters is illegal")
        else:
            self.ctr = cnt

class Base:
    """
    Base class for all abstract components. It provides

    Functionality for ID counting:
        - an automatic ``GLOBAL_ID`` unique during the lifetime of the program
        - a ``PREV_GLOBAL_ID`` for re-identification after save & load,
          automatically set in ``load()``

    Serialization:
        - the ``data()`` method gets reimplemented by subclasses to serialize
        - the ``load()`` method gets reimplemented by subclasses to deserialize
        - the complete_data_function attribute can be set to a function which\\
          extends the data dict with additional information, which is useful
          when ryvencore is embedded e.g. in a frontend
    """

    # all abstract components have a global ID
    _global_id_ctr = IDCtr()

    # TODO: this produces a memory leak, because the objects are never removed
    #  from the dict. It shouldn't be a problem as long as PREF_GLOBAL_ID is
    #  only used for loading.
    _prev_id_objs = {}

    @classmethod
    def obj_from_prev_id(cls, prev_id: int):
        """returns the object with the given previous id"""
        return cls._prev_id_objs.get(prev_id)

    def __init__(self):
        self.GLOBAL_ID = self._global_id_ctr.count()
        self.PREV_GLOBAL_ID = None

    """
    
    CUSTOM DATA
    
    """

    # this can be conveniently set to another function by the host to implement
    # adding additional (e.g. frontend-related) information to the data dict
    complete_data_function = complete_data

    def data(self) -> dict:
        """converts the object to a JSON compatible dict for serialization"""
        return {'GID': self.GLOBAL_ID}

    def complete_data(self, data: dict) -> data:
        return Base.complete_data_function(data)

    def load(self, data: dict):
        """recreate the object from the data dict returned by data()"""
        if dict is not None:
            self.PREV_GLOBAL_ID = data['GID']
            self._prev_id_objs[self.PREV_GLOBAL_ID] = self



"""
This file defines the Data type, which must be used to pass data between nodes.
It should be subclassed to define custom data types. In particular, serialization
and deserialization must be implemented for each respective type. Types that are
pickle serializable by default can be used without subclassing (``Data(my_data)``).
"""


class Data:
    """
    Base class for data objects.

    Subclass this class and implement serialization and deserialization accordingly
    to send data to other nodes.

    In case of large data sets being shared, you might want to leave serialization
    empty, which means the graph will not enter the same state when you reload it,
    which is fine as long as your nodes are built appropriately e.g. such that you can
    quickly regenerate that state by updating the root node.

    Be careful with nodes built for sharing complex output data over multiple outputs.
    Refer to python's object referencing rules. In particular, if you share data object
    `D` with successor nodes `N1` and `N2`, and `N1` changes `D`, then `N2` will also
    see the change.
    To avoid this you might want to make sure to copy `D` once it's consumed for the
    second time, which you can also conveniently implement in your ``Data`` class.
    """

    def __init__(self, value=None, load_from=None):
        if load_from is not None:
            self.set_data(load_from)
        else:
            self.payload = value

    def get_data(self):
        """
        Transform the data object to a ``pickle``serializable object.
        """
        return self.payload

    def set_data(self, data):
        """
        Deserialize the data object from the dict created in ``serialize()``.
        """
        self.payload = data



import importlib
import glob
from typing import List, Dict

from .Base import Base, Event
from .Flow import Flow
from .InfoMsgs import InfoMsgs
from .utils import pkg_path, load_from_file
from .Node import Node


class Session(Base):
    """
    The Session is the top level interface to your project. It mainly manages flows and registered nodes, and
    provides methods for serialization and deserialization of the project.
    """

    def __init__(
            self,
            gui: bool = False,
    ):
        Base.__init__(self)

        # events
        self.new_flow_created = Event(Flow)
        self.flow_renamed = Event(Flow)
        self.flow_deleted = Event(Flow)

        # ATTRIBUTES
        self.addons = {}
        self.flows: [Flow] = []
        self.nodes = set()  # list of node CLASSES
        self.invisible_nodes = set()
        self.gui: bool = gui
        self.init_data = None

        self.load_addons(pkg_path('addons/default/'))
        self.load_addons(pkg_path('addons/'))


    def load_addons(self, location: str):
        """
        Loads all addons from the given location. ``location`` can be an absolute path to any readable directory.
        """

        # discover all top-level modules in the given location
        addons = filter(lambda p: not p.endswith('__init__.py'), glob.glob(location + '/*.py'))

        for path in addons:
            # extract 'addon' object from module
            addon, = load_from_file(path, ['addon'])

            if addon is None:
                continue

            # register addon
            modname = path.split('/')[-1][:-3]
            self.addons[modname] = addon

            addon.register(self)
            setattr(Node, addon.name, addon)


    def register_nodes(self, node_classes: List):
        """Registers a list of Nodes which then become available in the flows"""

        for n in node_classes:
            self.register_node(n)


    def register_node(self, node_class):
        """Registers a single Node which then becomes available in the flows"""

        # build node class identifier
        node_class.build_identifier()

        self.nodes.add(node_class)


    def unregister_node(self, node_class):
        """Unregisters a Node which will then be removed from the available list.
        Existing instances won't be affected."""

        self.nodes.remove(node_class)


    def all_node_objects(self) -> List:
        """Returns a list of all Node objects instantiated in any flow"""

        nodes = []
        for s in self.flows:
            for n in s.flow.nodes:
                nodes.append(n)
        return nodes


    def create_flow(self, title: str = None, data: Dict = None) -> Flow:
        """Creates and returns a new flow.
        If data is provided the title parameter will be ignored."""

        flow = Flow(session=self, title=title)
        self.flows.append(flow)

        if data:
            flow.load(data)

        self.new_flow_created.emit(flow)

        return flow


    def rename_flow(self, flow: Flow, title: str) -> bool:
        """Renames an existing flow and returns success boolean"""

        success = False

        if self.flow_title_valid(title):
            flow.title = title
            success = True

        self.flow_renamed.emit(flow)

        return success


    def flow_title_valid(self, title: str) -> bool:
        """Checks whether a considered title for a new flow is valid (unique) or not"""

        if len(title) == 0:
            return False
        for s in self.flows:
            if s.title == title:
                return False

        return True


    def delete_flow(self, flow: Flow):
        """Removes an existing flow."""

        self.flows.remove(flow)

        self.flow_deleted.emit(flow)


    def info_messenger(self):
        """Returns a reference to InfoMsgs to print info data"""

        return InfoMsgs


    def load(self, data: Dict) -> List[Flow]:
        """Loads a project and raises an exception if required nodes are missing"""
        super().load(data)

        self.init_data = data

        # load flows
        new_flows = []

        #   backward compatibility
        if 'scripts' in data:
            flows_data = {
                title: script_data['flow']
                for title, script_data in data['scripts'].items()
            }
        else:
            flows_data = data['flows']

        for fd in flows_data:
            new_flows.append(self.create_flow(data=fd))

        # load addons
        for name, addon_data in data['addons'].items():
            if name in self.addons:
                self.addons[name].set_state(addon_data)
            else:
                print(f'found missing addon: {name}; attempting to load anyway')

        return new_flows

    def serialize(self):
        """Returns the project as JSON compatible dict to be saved and loaded again using load()"""

        return self.complete_data(self.data())


    def data(self) -> dict:
        d = super().data()
        d.update({
            'flows': [
                s.data() for s in self.flows
            ],
            'addons': {
                name: addon.get_state() for name, addon in self.addons.items()
            }
        })
        return d



"""
The flow executors are responsible for executing the flow. They have access to
the flow as well as the nodes' internals and are able to perform optimizations.
"""
from .Data import Data
from .RC import FlowAlg


"""
graph_adjacency = {
    node_output: [node_input],    
    node_input:  node_output | None
}
"""


class FlowExecutor:
    """
    Base class for special flow execution algorithms.
    """

    def __init__(self, flow):
        self.flow = flow
        self.flow_changed = True
        self.graph = self.flow.graph_adj
        self.graph_rev = self.flow.graph_adj_rev

    # Node.update() =>
    def update_node(self, node, inp):
        pass

    # Node.input() =>
    def input(self, node, index):
        pass

    # Node.set_output_val() =>
    def set_output_val(self, node, index, val):
        pass

    # Node.exec_output() =>
    def exec_output(self, node, index):
        pass

    def conn_added(self, out, inp, silent=False):
        pass

    def conn_removed(self, out, inp):
        pass


class DataFlowNaive(FlowExecutor):
    """
    The naive implementation of data-flow execution. Naive meaning setting a node output
    leads to an immediate update in all successors consecutively. No runtime optimization
    if performed, and some types of graphs can run really slow here, especially if they
    include "diamonds".

    Assumptions for the graph:
    - no non-terminating feedback loops
    """

    # Node.update() =>
    def update_node(self, node, inp):
        try:
            node.update_event(inp)
        except Exception as e:
            node.update_error(e)

    # Node.input() =>
    def input(self, node, index):
        inp = node.inputs[index]
        conn_out = self.graph_rev[inp]

        if conn_out:
            return conn_out.val
        else:
            return None

    # Node.set_output_val() =>
    def set_output_val(self, node, index, data):
        out = node.outputs[index]
        if not out.type_ == 'data':
            return
        out.val = data

        for inp in self.graph[out]:
            inp.node.update(inp=inp.node.inputs.index(inp))

    # Node.exec_output() =>
    def exec_output(self, node, index):
        out = node.outputs[index]
        if not out.type_ == 'exec':
            return

        for inp in self.graph[out]:
            inp.node.update(inp=inp.node.inputs.index(inp))

    def conn_added(self, out, inp, silent=False):
        if not silent:
            # update input
            inp.node.update(inp=inp.node.inputs.index(inp))

    def conn_removed(self, out, inp):
        # update input
        inp.node.update(inp=inp.node.inputs.index(inp))


class DataFlowOptimized(DataFlowNaive):
    """
    A special flow executor which implements some node functions to optimise flow execution.
    Whenever a new execution is invoked somewhere (some node or output is updated), it
    analyses the graph's connected component (of successors) where the execution was invoked
    and creates a few data structures to reverse engineer how many input
    updates every node possibly receives in this execution. A node's outputs are
    propagated once no input can still receive new data from a predecessor node.
    Therefore, while a node gets updated every time an input receives some data,
    every OUTPUT is only updated ONCE.
    This implies that every connection is activated at most once in an execution.
    This can result in asymptotic speedup in large data flows compared to normal data flow
    execution where any two executed branches which merge again in the future result in two
    complete executions of everything that comes after the merge, which quickly produces
    exponential performance issues.

    Assumptions for the graph:

    - no feedback loops
    """

    def __init__(self, flow):
        super().__init__(flow)

        self.output_updated = {}
        self.waiting_count = {}
        self.node_waiting = {}
        self.num_conns_from_predecessors = None
        self.last_execution_root = None     # for reuse when a same execution is invoked many times consecutively
        self.execution_root = None          # can be Node or NodeOutput
        self.execution_root_node = None     # the updated Node or the updated NodeOutput's Node

    # NODE FUNCTIONS

    # Node.update() =>
    def update_node(self, node, inp=-1):
        if self.execution_root_node is None:  # execution starter!
            self.start_execution(root_node=node)
            self.invoke_node_update_event(node, inp)
            self.propagate_outputs(node)
            self.stop_execution()
        else:
            self.invoke_node_update_event(node, inp)

    # Node.input() =>
    #   DataFlowNative.input(node, index)

    # Node.set_output_val() =>
    def set_output_val(self, node, index, data):
        out = node.outputs[index]

        if self.execution_root_node is None:  # execution starter!
            self.start_execution(root_output=out)

            out.val = data
            self.output_updated[out] = True
            self.propagate_output(out)

            self.stop_execution()

        else:

            if not self.node_waiting[out.node]:
                # the output's node might not be part of the analyzed graph!
                # in this case we immediately push the value
                # there are other possible solutions to this, including running
                # a new execution analysis of this graph here

                super().set_output_val(node, index, data)

            else:
                out.val = data
                self.output_updated[out] = True

    # Node.exec_output() =>
    def exec_output(self, node, index):
        # rudimentary exec support also in data flows

        out = node.outputs[index]

        if self.execution_root_node is None:  # execution starter!
            self.start_execution(root_output=out)

            self.output_updated[out] = True
            self.propagate_output(out)

            self.stop_execution()

        else:
            self.output_updated[out] = True

    """
    
    Helper methods
    
    """

    def start_execution(self, root_node=None, root_output=None):

        # reset cached output values
        self.output_updated = {}
        for n in self.flow.nodes:
            for out in n.outputs:
                self.output_updated[out] = False

        if root_node is not None:
            self.execution_root = root_node
            self.execution_root_node = root_node
            self.waiting_count = self.generate_waiting_count(root_node=root_node)

        elif root_output is not None:
            self.execution_root = root_output
            self.execution_root_node = root_output.node
            self.waiting_count = self.generate_waiting_count(root_output=root_output)

    def stop_execution(self):
        self.execution_root_node = None
        self.last_execution_root = self.execution_root
        self.execution_root = None

    def generate_waiting_count(self, root_node=None, root_output=None):
        if not self.flow_changed and self.execution_root is self.last_execution_root:
            return self.num_conns_from_predecessors.copy()
        self.flow_changed = False

        nodes = self.flow.nodes
        node_successors = self.flow.node_successors

        # DP TABLE
        self.num_conns_from_predecessors = {
            n: 0
            for n in nodes
        }

        successors = set()
        visited = {
            n: False
            for n in nodes
        }

        # BC
        if root_node is not None:
            successors.add(root_node)

        elif root_output is not None:
            for inp in self.graph[root_output]:
                connected_node = inp.node
                self.num_conns_from_predecessors[connected_node] += 1
                successors.add(connected_node)

        # ITERATION
        while len(successors) > 0:
            n = successors.pop()
            if visited[n]:
                continue

            for s in node_successors[n]:
                self.num_conns_from_predecessors[s] += 1
                successors.add(s)
            visited[n] = True

        self.node_waiting = visited

        return self.num_conns_from_predecessors.copy()

    def invoke_node_update_event(self, node, inp):
        super().update_node(node, inp)

    def decrease_wait(self, node):
        """decreases the wait count of the node;
        if the count reaches zero, which means there is no other input waiting for data,
        the output values get propagated"""

        self.waiting_count[node] -= 1
        if self.waiting_count[node] == 0:
            self.propagate_outputs(node)

    def propagate_outputs(self, node):
        """propagates all outputs of node"""

        for out in node.outputs:
            self.propagate_output(out)

    def propagate_output(self, out):
        """pushes an output's value to successors if it has been changed in the execution"""

        if self.output_updated[out]:
            # same procedure for data and exec connections
            for inp in self.graph[out]:
                inp.node.update(inp=inp.node.inputs.index(inp))

        # decrease wait count of successors
        for inp in self.graph[out]:
            self.decrease_wait(inp.node)


class ExecFlowNaive(FlowExecutor):
    """
    ...
    """

    def __init__(self, flow):
        super().__init__(flow)

        # all the nodes currently updating because of an output data request
        # used to prevent redundant predecessor updates during the update
        # of a single successor
        self.updated_nodes = None

    # Node.update() = >
    def update_node(self, node, inp):
        if inp != -1 and node.inputs[inp].type_ == 'data':
            return

        execution_starter = self.updated_nodes is None

        if execution_starter:
            self.updated_nodes = {node}
        else:
            self.updated_nodes.add(node)

        try:
            node.update_event(inp)
        except Exception as e:
            node.update_error(e)

        if execution_starter:
            self.updated_nodes = None

    # Node.input() =>
    def input(self, node, index):
        inp = node.inputs[index]
        out = self.graph_rev[inp]
        if out:
            n = out.node
            if n not in self.updated_nodes:
                n.update(-1)

            return out.val
        else:
            return None

    # Node.set_output_val() =>
    def set_output_val(self, node, index, data):
        out = node.outputs[index]
        out.val = data

    # Node.exec_output() =>
    def exec_output(self, node, index):
        for inp in self.graph[node.outputs[index]]:
            inp.node.update(inp.node.inputs.index(inp))


def executor_from_flow_alg(algorithm: FlowAlg):
    if algorithm == FlowAlg.DATA:
        return DataFlowNaive
    if algorithm == FlowAlg.DATA_OPT:
        return DataFlowOptimized
    if algorithm == FlowAlg.EXEC:
        return ExecFlowNaive



"""
This module defines a simple add-on system to extend ryvencore's functionalities.
Some default add-ons are provided in the add-ons.default package, and additional add-ons
can be added and registered in the Session.

An add-on

- has a name, a version, a description
- is session-local, not flow-local (but you can of course implement per-flow functionality)  
- manages its own state (in particular ``get_state()`` and ``set_state()``)
- will be accessible by nodes as attribute (self.add-on_name in the node)

Add-on access is blocked during loading, so nodes should not access any add-ons in set_data(). 
This prevents inconsistent states. Nodes are loaded first, then the add-ons. 
Therefore, the add-on should be sufficiently isolated and self-contained.

To define a custom add-on you need to subclass the ``AddOn`` class in its own module,
instantiate it into a top module level variable ``addon`` (``addon = YourAddon()``),
and put the module into an add-on directory. See ``Session.register_addon`` and
see ``ryvencore.addons.default`` for examples.
"""

from ryvencore.Base import Base


class AddOn(Base):
    name = ''
    version = ''
    description = ''

    def register(self, session):
        """
        Called when the add-on is registered with a session.
        """
        self.session = session

    def _on_node_created(self, flow, node):
        """
        Called when a node is created. This happens only once, whereas
        a node can be added and removed multiple times, see
        on_node_added() and
        on_node_removed().
        """
        pass

    def _on_node_added(self, flow, node):
        """
        Called when a node is added to a flow.
        """
        pass

    def _on_node_removed(self, flow, node):
        """
        Called when a node is removed from a flow.
        """
        pass

    def _extend_node_data(self, node, data: dict):
        """
        Extend the node data dict with additional add-on-related data.
        """
        pass

    # def _extend_flow_data(self, flow, data: dict):
    #     """
    #     Extend the flow data dict with additional add-on-related data.
    #     """
    #     pass
    #
    # def _extend_session_data(self, data: dict):
    #     """
    #     Extend the session data dict with additional add-on-related data.
    #     """
    #     pass

    def get_state(self) -> dict:
        """
        Return the state of the add-on as JSON-compatible a dict.
        """
        return {}

    def set_state(self, state: dict):
        """
        Set the state of the add-on from the dict generated in get_state().
        """
        pass



"""
This module defines dtypes, definitions of declarative types which can be conveniently used for
data inputs, and an add-on for those.

These dtypes only store some information which might be used by the nodes, or by a frontend, etc.
There is currently no type checking or something like that implemented, but might be added
optionally in the future.

This list may grow significantly over time.
"""

from typing import Dict as t_Dict, List as t_List


class DType:
    def __init__(self, default, bounds: tuple = None, doc: str = "", _load_state=None):

        self.default = default
        self.val = self.default
        self.doc = doc
        self.bounds = bounds

        if _load_state:
            self.set_state(_load_state)

        self._data = ['default', 'val', 'doc', 'bounds']

    def __str__(self):
        return 'DType.'+self.__class__.__name__

    @staticmethod
    def from_str(s):
        for DTypeClass in dtypes:
            if s == 'DType.'+DTypeClass.__name__:
                return DTypeClass

        return None

    def add_data(self, *attr_names):
        self._data += list(attr_names)

    def get_state(self) -> dict:
        return {
            name: getattr(self, name)
            for name in self._data
        }

    def set_state(self, data: dict):
        for name, val in data.items():
            setattr(self, name, val)


class Data(DType):
    """Any kind of data represented by some evaluated text input"""
    def __init__(self, default=None, size: str = 'm', doc: str = "", _load_state=None):
        """
        size: 's' / 'm' / 'l'
        """
        self.size = size
        super().__init__(default=default, doc=doc, _load_state=_load_state)
        self.add_data('size')


class Integer(DType):
    def __init__(self, default: int = 0, bounds: tuple = None, doc: str = "", _load_state=None):
        super().__init__(default=default, bounds=bounds, doc=doc, _load_state=_load_state)


class Float(DType):
    def __init__(self, default: float = 0.0, bounds: tuple = None, decimals: int = 10, doc: str = "", _load_state=None):
        self.decimals = decimals
        super().__init__(default=default, bounds=bounds, doc=doc, _load_state=_load_state)
        self.add_data('decimals')


class Boolean(DType):
    def __init__(self, default: bool = False, doc: str = "", _load_state=None):
        super().__init__(default=default, doc=doc, _load_state=_load_state)


class Char(DType):
    def __init__(self, default: chr = '', doc: str = "", _load_state=None):
        super().__init__(default=default, doc=doc, _load_state=_load_state)


class String(DType):
    def __init__(self, default: str = "", size: str = 'm', doc: str = "", _load_state=None):
        """
        size: 's' / 'm' / 'l'
        """
        self.size = size
        super().__init__(default=default, doc=doc, _load_state=_load_state)
        self.add_data('size')


class Choice(DType):
    def __init__(self, default=None, items: t_List = [], doc: str = "", _load_state=None):
        self.items = items
        super().__init__(default=default, doc=doc, _load_state=_load_state)
        self.add_data('items')


class List(DType):
    def __init__(self, default: t_List = [], doc: str = "", _load_state=None):
        super().__init__(default=default, doc=doc, _load_state=_load_state)


class Date(DType):
    ...


class Time(DType):
    ...


class Color(DType):
    ...


class Range(DType):
    ...


# dtypes = [Data, Integer, Float, Boolean, Char, String, Choice, List]



'''

Node: ...

    def create_input_dt(self, dtype: DType, label: str = '', add_data={}, insert: int = None):
        """Creates and adds a new data input with a DType"""
        # InfoMsgs.write('create_input called')

        inp = NodeInput(
            node=self,
            type_='data',
            label_str=label,
            dtype=dtype,
            add_data=add_data,
        )

        if insert is not None:
            self.inputs.insert(insert, inp)
        else:
            self.inputs.append(inp)


NodePort.data(): ...

        if self.dtype:
            data['dtype'] = str(self.dtype)
            data['dtype state'] = serialize(self.dtype.get_state())
'''

from ryvencore import AddOn
from ryvencore import NodeInputType


class DtypesAddon(AddOn):
    """
    An add-on that adds the ability to create data inputs with a DType.
    A DType simply holds some information about the data supposed to be
    processed at the respective input, but does not perform any typechecks.

    Currently, DType inputs can only be created dynamically through
    ``DTypes.create_dtype_input(self, ...)``, not in ``init_inputs``. This
    might change once we found a way of communicating the DType information
    back to the add-on when node inputs are built from ``init_inputs``,
    which currently isn't easily possible.
    """

    name = 'DTypes'
    version = '0.0.1'

    Data = Data
    Integer = Integer
    Float = Float
    Boolean = Boolean
    Char = Char
    String = String
    Choice = Choice
    List = List

    class NodeInputType(NodeInputType):
        def __init__(self, dtype=None, *args, **kwargs):
            super().__init__(*args, **kwargs)

            self.dtype = dtype

    def __init__(self):
        super().__init__()

        self.dtype_inputs = {}  # {NodeInput: DType}
        self.creating_dtype_input = False

    def create_dtype_input(self, node, dtype, *args, **kwargs):
        """
        Dynamically creates a new input and assigns the given dtype.
        """

        self.creating_dtype_input = True
        inp = node.create_input(*args, **kwargs)
        self.creating_dtype_input = False

        # dtypes are only on data inputs allowed
        if inp.type_ != 'data':
            node.delete_input(node.inputs.index(inp))
            return

        self.dtype_inputs[inp] = dtype

    def _on_node_created(self, flow, node):
        """
        Restores the node's dtypes.
        """

        for i, inp in enumerate(node.inputs):
            if inp.type_ == 'data' and 'dtype' in node.init_data['inputs'][i]:
                dtype = DType.from_str(node.init_data['inputs'][i]['dtype'])
                dtype.set_state(node.init_data['inputs'][i]['dtype']['state'])
                self.dtype_inputs[inp] = dtype

    def _extend_node_data(self, node, data: dict):
        for i, inp in enumerate(node.inputs):
            if inp in self.dtype_inputs:
                data['inputs'][i]['dtype'] = {
                    'type': str(self.dtype_inputs[inp]),
                    'state': self.dtype_inputs[inp].get_state(),
                }

# addon = DtypesAddon()



from logging import Logger as PyLogger
from typing import Optional

from ryvencore.AddOn import AddOn


class Logger(PyLogger):

    def __init__(self, *args, **kwargs):
        PyLogger.__init__(self, *args, **kwargs)
    #     # events
    #     self.sig_enabled = Event()
    #     self.sig_disabled = Event()

    def enable(self):
        # self.sig_enabled.emit()
        pass

    def disable(self):
        # self.sig_disabled.emit()
        pass


class LoggingAddon(AddOn):
    """
    This addon implements very basic some logging functionality.

    It provides an API to create and delete loggers that are owned
    by a particular node. The logger gets enabled/disabled
    automatically when the owning node is added to/removed from
    the flow.

    Ownership might eventually be expanded to any component that
    preserves its global ID throughout save and load.

    The contents of logs are currently not preserved. If a log's
    content should be preserved, it should be saved in a file.

    Refer to Python's logging module documentation.
    """

    name = 'Logging'
    version = '0.0.1'

    def __init__(self):
        super().__init__()

        # logger_created = Event(Logger)

        self.loggers = {}   # {Node: {name: Logger}}

    def new_logger(self, node, title: str) -> Optional[Logger]:
        """
        Creates a new logger owned by the node, returns None if
        one with the given name already exists.
        """

        if not self._node_is_registered(node):
            self.loggers[node] = {}

        elif title in self.loggers[node]:
            return None

        logger = Logger(name=title)
        self.loggers[node][title] = logger
        # self.logger_created.emit(logger)
        return logger

    def _on_node_created(self, flow, node):
        if node.init_data and 'Logging' in node.init_data:
            for title in node.init_data['Logging']['loggers']:
                self.new_logger(node, title)
                # in case the node already created the logger,
                # new_logger() will have no effect

    def _node_is_registered(self, node):
        return node in self.loggers

    def _on_node_added(self, flow, node):
        if not self._node_is_registered(node):
            return

        # enable the node's loggers
        for logger in self.loggers[node].values():
            logger.enable()

    def _on_node_removed(self, flow, node):
        if not self._node_is_registered(node):
            return

        # disable the node's loggers
        for logger in self.loggers[node].values():
            logger.disable()

    def _extend_node_data(self, node, data: dict):
        if not self._node_is_registered(node):
            return

        data['Logging'] = {
            'loggers': [name for name in self.loggers[node].keys()]
        }

addon = LoggingAddon()



"""
deprecated and now unused
"""

from .Base import Base, Event

from .InfoMsgs import InfoMsgs


class Connection(Base):
    """
    The base class for both types of connections. All data is transmitted through a connection from an output
    port to some connected input port.
    """

    def __init__(self, params):
        Base.__init__(self)

        self.activated = Event(object)  # TODO: connections are not activated anymore

        self.out, self.inp, self.flow = params


class ExecConnection(Connection):
    pass


class DataConnection(Connection):

    def __init__(self, params):
        super().__init__(params)

        self.data = None



from .InfoMsgs import InfoMsgs
from .RC import *
from .Session import Session
from .Flow import Flow
from .Data import Data
from .AddOn import AddOn
from .Node import Node
from .NodePortType import NodeInputType, NodeOutputType
from .Connection import DataConnection, ExecConnection
from .utils import serialize, deserialize



from .Base import Base, Event
from .Data import Data
from .FlowExecutor import DataFlowNaive, DataFlowOptimized, FlowExecutor, executor_from_flow_alg
from .Node import Node
from .NodePort import NodePort
from .RC import FlowAlg, PortObjPos
from .utils import node_from_identifier, serialize, deserialize
from typing import List, Dict, Optional, Tuple


class Flow(Base):
    """
    Manages all abstract flow components (nodes, connections) and includes implementations for editing.
    """

    def __init__(self, session, title):
        Base.__init__(self)

        # events
        self.node_added = Event(Node)
        self.node_removed = Event(Node)
        self.connection_added = Event((NodePort, NodePort))        # Event(Connection)
        self.connection_removed = Event((NodePort, NodePort))      # Event (Connection)

        self.connection_request_valid = Event(bool)
        self.nodes_created_from_data = Event(list)
        self.connections_created_from_data = Event(list)

        self.algorithm_mode_changed = Event(str)

        # general attributes
        self.session = session
        self.title = title
        self.nodes: [Node] = []

        self.node_successors = {}   # additional data structure for executors
        self.graph_adj = {}         # directed adjacency list relating node ports
        self.graph_adj_rev = {}     # reverse adjacency; reverse of graph_adj

        self.alg_mode = FlowAlg.DATA
        self.executor: FlowExecutor = executor_from_flow_alg(self.alg_mode)(self)

    def load(self, data: dict):
        """Loading a flow from data as previously returned by data()"""
        super().load(data)

        # set algorithm mode
        self.alg_mode = FlowAlg.from_str(data['algorithm mode'])

        # build flow

        new_nodes = self.create_nodes_from_data(data['nodes'])

        self.set_output_values_from_data(new_nodes, data['output data'])

        self.connect_nodes_from_data(new_nodes, data['connections'])


    def create_nodes_from_data(self, nodes_data: List):
        """create nodes from nodes_data as previously returned by data()"""

        nodes = []

        for n_c in nodes_data:

            # find class
            node_class = node_from_identifier(
                n_c['identifier'],
                self.session.nodes.union(self.session.invisible_nodes)
            )

            node = self.create_node(node_class, n_c)
            nodes.append(node)

        self.nodes_created_from_data.emit(nodes)

        return nodes


    def set_output_values_from_data(self, nodes: List[Node], data: List):
        for d in data:
            indices = d['dependent node outputs']
            indices_paired = zip(indices[0::2], indices[1::2])
            for node_index, output_index in indices_paired:
                nodes[node_index].outputs[output_index].val = \
                    Data(load_from=deserialize(d['data']))


    def create_node(self, node_class, data=None):
        """Creates, adds and returns a new node object"""

        node = node_class((self, self.session, data))
        node.initialize()
        self.add_node(node)

        # notify addons
        for addon in self.session.addons.values():
            addon._on_node_created(self, node)

        return node


    def add_node(self, node: Node):
        """Stores a node object and causes the node's place_event()"""

        self.nodes.append(node)

        self.node_successors[node] = []
        for out in node.outputs:
            self.graph_adj[out] = []
        for inp in node.inputs:
            self.graph_adj_rev[inp] = None

        node.after_placement()
        self.flow_changed()

        # notify addons
        for addon in self.session.addons.values():
            addon._on_node_added(self, node)

        self.node_added.emit(node)


    def node_view_placed(self, node: Node):
        """Triggered after the node's GUI content has been fully initialized"""

        node.view_place_event()


    def remove_node(self, node: Node):
        """Removes a node from internal list without deleting it"""

        node.prepare_removal()
        self.nodes.remove(node)

        del self.node_successors[node]
        for out in node.outputs:
            del self.graph_adj[out]
        for inp in node.inputs:
            del self.graph_adj_rev[inp]

        self.flow_changed()

        # notify addons
        for addon in self.session.addons.values():
            addon._on_node_removed(self, node)

        self.node_removed.emit(node)


    def connect_nodes_from_data(self, nodes: List[Node], data: List):
        connections = []

        for c in data:

            c_parent_node_index = c['parent node index']
            c_connected_node_index = c['connected node']
            c_output_port_index = c['output port index']
            c_connected_input_port_index = c['connected input port index']

            if c_connected_node_index is not None:  # which can be the case when pasting
                parent_node = nodes[c_parent_node_index]
                connected_node = nodes[c_connected_node_index]

                connections.append(
                    self.connect_nodes(
                        parent_node.outputs[c_output_port_index],
                        connected_node.inputs[c_connected_input_port_index],
                        silent=True
                    ))

        self.connections_created_from_data.emit(connections)

        return connections


    def check_connection_validity(self, p1: NodePort, p2: NodePort) -> bool:
        """Checks whether a considered connect action is legal"""

        valid = True

        if p1.node == p2.node:
            valid = False

        if p1.io_pos == p2.io_pos or p1.type_ != p2.type_:
            valid = False

        self.connection_request_valid.emit(valid)

        return valid


    def connect_nodes(self, p1: NodePort, p2: NodePort, silent: bool = False) -> Optional[Tuple[NodePort, NodePort]]:
        """Connects nodes or disconnects them if they are already connected"""

        if not self.check_connection_validity(p1, p2):
            return None

        out = p1
        inp = p2
        if out.io_pos == PortObjPos.INPUT:
            out, inp = inp, out

        if inp in self.graph_adj[out]:
            # disconnect
            self.remove_connection((out, inp))
            return None

        self.add_connection((out, inp), silent=silent)

        return out, inp


    def add_connection(self, c: Tuple[NodePort, NodePort], silent: bool = False):
        """Adds a connection object"""

        out, inp = c

        self.graph_adj[out].append(inp)
        self.graph_adj_rev[inp] = out

        self.node_successors[out.node].append(inp.node)
        self.flow_changed()


        self.executor.conn_added(out, inp, silent=silent)

        self.connection_added.emit((out, inp))


    def remove_connection(self, c: Tuple[NodePort, NodePort]):
        """Removes a connection object without deleting it"""

        out, inp = c

        self.graph_adj[out].remove(inp)
        self.graph_adj_rev[inp] = None

        self.node_successors[out.node].remove(inp.node)
        self.flow_changed()

        self.executor.conn_removed(out, inp)
#
        self.connection_removed.emit((out, inp))


    def connected_inputs(self, out: NodePort) -> List[NodePort]:
        return self.graph_adj[out]


    def connected_output(self, inp: NodePort) -> Optional[NodePort]:
        return self.graph_adj_rev[inp]


    def algorithm_mode(self) -> str:
        """Returns the current algorithm mode of the flow as string"""

        return FlowAlg.str(self.alg_mode)


    def set_algorithm_mode(self, mode: str):
        """Sets the algorithm mode of the flow, possible values are 'data' and 'exec'"""

        new_alg_mode = FlowAlg.from_str(mode)
        if new_alg_mode is None:
            return False

        self.executor = executor_from_flow_alg(new_alg_mode)(self)
        self.alg_mode = new_alg_mode
        self.algorithm_mode_changed.emit(self.algorithm_mode())

        return True


    def flow_changed(self):
        self.executor.flow_changed = True


    def data(self) -> dict:
        """Returns a dictionary containing all data of the flow"""
        d = super().data()
        d.update({
            'algorithm mode': FlowAlg.str(self.alg_mode),
            'nodes': self.gen_nodes_data(self.nodes),
            'connections': self.gen_conns_data(self.nodes),
            'output data': self.gen_output_data(self.nodes),
        })
        return d


    def gen_nodes_data(self, nodes: List[Node]) -> List[dict]:
        """Returns the data dicts of the nodes given"""

        return [n.data() for n in nodes]


    def gen_conns_data(self, nodes: List[Node]) -> List[dict]:
        """Generates the connections data between and relative to the nodes passed"""

        # notice that this is intentionally not part of Connection, because connection data
        # is generated always for a specific set of nodes (like all currently selected ones)
        # and the data dict therefore has the refer to the indices of the nodes in the nodes list

        data = []
        for i, n in enumerate(nodes):
            for j, out in enumerate(n.outputs):
                for inp in self.graph_adj[out]:
                    if inp.node in nodes:
                        data.append({
                            'parent node index': i,
                            'output port index': j,
                            'connected node': nodes.index(inp.node),
                            'connected input port index': inp.node.inputs.index(inp),
                        })

        return data


    def gen_output_data(self, nodes: List[Node]) -> List[Dict]:
        """Serializes output data of the nodes"""

        outputs_data = {}

        for i_n, n in enumerate(nodes):
            for i_o, out in enumerate(n.outputs):
                d = out.val
                if isinstance(d, Data) and d not in outputs_data:
                    outputs_data[d] = {
                        'data': serialize(d.get_data()),
                        'dependent node outputs': [i_n, i_o],
                    }
                elif isinstance(d, Data):
                    outputs_data[d]['dependent node outputs'] += [i_n, i_o]

        return list(outputs_data.values())



from typing import Optional

from ryvencore import Node
from ryvencore.Base import Base, Event
from ryvencore.utils import serialize, deserialize
from ryvencore import AddOn



class Variable:
    """Implementation of flow variables"""

    def __init__(self, addon, flow, name='', val=None, data=None):
        self.addon = addon
        self.flow = flow
        self.name = name
        self.val = val

        if data and 'serialized' in data.keys():
            self.val = deserialize(data['serialized'])

    def get(self):
        """
        Returns the value of the variable
        """
        return self.val

    def set(self, val):
        """
        Sets the value of the variable
        """
        self.val = val
        self.addon._var_updated(self.flow, self.name)

    def serialize(self):
        return serialize(self.val)




class VarsAddon(AddOn):
    """
    This addon provides a simple variable system.

    It provides an API to create Variable objects which can wrap any Python object.

    Nodes can subscribe to variable names with a callback that is executed once a
    variable with that name changes or is created. The callback must be a method of
    the node, so the subscription can be re-established on loading.

    This way nodes can react to changes of data and non-trivial data-flow is introduced,
    meaning that data dependencies are determined also by variable subscriptions and not
    purely by the edges in the graph anymore. This can be useful, but it can also prevent
    optimization. Variables are flow-local.
    """

    name = 'Variables'
    version = '0.0.3'

    def __init__(self):
        AddOn.__init__(self)

        # {
        #   Flow: {
        #       'variable name': {
        #           'var': Variable,
        #           'subscriptions': [Node method]
        #        },
        # }
        self.flow_variables = {}

    def var_name_valid(self, flow, name: str) -> bool:
        """
        Checks if ``name`` is a valid variable identifier and hasn't been take yet.
        """

        return name.isidentifier() and not self._var_exists(flow, name)

    def create_var(self, flow, name: str, val=None, data=None) -> Optional[Variable]:
        """
        Creates and returns a new variable and None if the name isn't valid.
        """

        if flow not in self.flow_variables:
            self.flow_variables[flow] = {}

        if self.var_name_valid(flow, name):
            v = Variable(self, flow, name, val, data)
            self.flow_variables[flow][name] = {
                'var': v,
                'subscriptions': []
            }
            return v
        else:
            return None

    def delete_var(self, flow, name: str):
        """
        Deletes a variable and causes subscription update. Subscriptions are preserved.
        """
        if not self._var_exists(flow, name):
            return

        del self.flow_variables[flow][name]['var']

    def _var_exists(self, flow, name: str) -> bool:
        return flow in self.flow_variables and name in self.flow_variables[flow]

    def var(self, flow, name: str):
        """
        Returns the variable with the given name or None if it doesn't exist.
        """
        if not self._var_exists(flow, name):
            return None

        return self.flow_variables[flow][name]['var']

    def _var_updated(self, flow, name: str):
        """
        Called when a Variable object changes or when the var is created or deleted.
        """

        v = self.flow_variables[flow][name]['var']

        for (node, cb) in self.flow_variables[flow][name]['subscriptions']:
            cb(v.val)

    def subscribe(self, node: Node, name: str, callback):
        """
        Subscribe to a variable. ``callback`` must be a method of the node.
        """
        if not self._var_exists(node.flow, name):
            return

        self.flow_variables[node.flow][name]['subscriptions'].append((node, callback))

    def unsubscribe(self, node, name: str, callback):
        """
        Unsubscribe from a variable.
        """
        if not self._var_exists(node.flow, name):
            return

        self.flow_variables[node.flow][name]['subscriptions'].remove((node, callback))

    def _extend_node_data(self, node, data: dict):
        """
        Extends the node data with the variable subscriptions.
        """

        data['Variables'] = {
            'subscriptions': {
                name: cb.__name__
                for name, var in self.flow_variables[node.flow].items()
                for (n, cb) in var['subscriptions']
                if node == n
            }
        }

        if data['Variables']['subscriptions'] == {}:
            del data['Variables']

    def _on_node_created(self, flow, node):
        """
        Reconstruction of subscriptions.
        """
        if node.init_data and 'Variables' in node.init_data:
            for name, cb_name in node.init_data['Variables']['subscriptions'].items():
                self.subscribe(node, name, getattr(node, cb_name))

    def get_state(self) -> dict:
        return {
            f.GLOBAL_ID: {
                name: {
                    'serialized': var['var'].serialize()
                }
                for name, var in self.flow_variables[f].items()
            }
            for f in self.flow_variables.keys()
        }

    def set_state(self, state: dict):

        for pref_flow_id, variables in state.items():
            f = Base.obj_from_prev_id(pref_flow_id)

            # recreate variables
            for name, var in variables.items():
                if self._var_exists(f, name):
                    self.var(f, name).set(deserialize(var['serialized']))
                else:
                    self.create_var(f, name, data=var)


addon = VarsAddon()



"""A collection of useful functions used by different components."""

import base64
import importlib.util
import pickle
from os.path import dirname, abspath, join, basename
from typing import List, Tuple


def serialize(data) -> str:
    return base64.b64encode(pickle.dumps(data)).decode('ascii')


def deserialize(data):
    return pickle.loads(base64.b64decode(data))


def node_from_identifier(identifier: str, nodes: List):

    for nc in nodes:
        if nc.identifier == identifier:
            return nc
    else:  # couldn't find a node with this identifier => search for identifier_comp
        for nc in nodes:
            if identifier in nc.identifier_comp:
                return nc
        else:
            raise Exception(
                f'could not find node class with identifier \'{identifier}\'. '
                f'if you changed your node\'s class name, make sure to add the old '
                f'identifier to the identifier_comp list attribute to provide '
                f'backwards compatibility.'
            )

def pkg_path(subpath: str = None):
    """
    Returns the path to the installed package root directory, optionally with a relative sub-path appended.
    Notice that this returns the path to the ryvencore package (ryvencore/ryvencore/) not the repository (ryvencore/).
    """

    p = dirname(__file__)
    if subpath is not None:
        p = join(p, subpath)
    return abspath(p)


def load_from_file(file: str, comps: List[str]) -> Tuple:
    """
    Imports components with name in ``comps`` from a python module.
    """
    # https://stackoverflow.com/questions/67631/how-do-i-import-a-module-given-the-full-path

    name = basename(file).split('.')[0]
    spec = importlib.util.spec_from_file_location(name, file)
    importlib.util.module_from_spec(spec)
    mod = spec.loader.load_module(name)
    # using load_module(name) instead of exec_module(mod) here,
    # because exec_module() somehow then registers it as "built-in"
    # which is wrong and e.g. prevents inspect from parsing the source

    def get_comp(c):
        try:
            return getattr(mod, c)
        except AttributeError:
            return None

    return tuple([get_comp(c) for c in comps])
